#import "WaveformController.h"
#import "WaveformChecker.h"
#import "WaveformViewController.h"
#import "WaveformView.h"
#import <AudioToolbox/AudioToolbox.h>

@implementation WaveformController

- (id)init {
    self = [super init];
    if(self) {
        waveformViewControllers = [[NSMutableArray alloc] initWithCapacity:5];
    }
    return self;
}

- (void)awakeFromNib {
    WaveformViewController *wvc;
    int i;

    [waveformChecker setDelegate:self];

    for(i=0;i<WFCLENGTH;++i)
        waveform[i] = 0;

    [waveformView setPointArray:waveform count:WFCLENGTH];

    for(i=0;i<5;++i) {
        wvc = [[[WaveformViewController alloc] init] autorelease];
        [waveformViewControllers addObject:wvc];
        [wvc setDelegate:self];
        [view addSubview:[wvc view]];
        [[wvc view] setFrameOrigin:NSMakePoint(([[wvc view] frame]).size.width*i, 0)];
    }

    [view setNeedsDisplay:YES];
}

- (void)waveformViewControllerDidModify:(WaveformViewController *)waveformViewController {
    int i;
    NSEnumerator *enm = [waveformViewControllers objectEnumerator];
    WaveformViewController *v;

    NSMutableString *s = [NSMutableString string];

    for(i=0;i<WFCLENGTH;++i)
        tempwaveform[i] = 0;

    while (v = [enm nextObject]) {
        if([v state] == NSOnState) {
            int wavecount = [v wavecount];
            double gain = [v gain];
            double phase = [v phase];

            if(![s isEqual:@""])
                [s appendString:@" + "];
            
            switch ([v shape]) {
                case WFSine:
                    for(i=0;i<WFCLENGTH;++i) {
                        double currentphase = (double)i/(double)WFCLENGTH*2*M_PI + phase;
                        currentphase *= wavecount;
                        while(currentphase > 2*M_PI)
                            currentphase -= 2*M_PI;
                        tempwaveform[i] += WFSIN(currentphase)*gain;
                    }
                    [s appendFormat:@"SIN"];
                    break;
                case WFRect:
                    for(i=0;i<WFCLENGTH;++i) {
                        double currentphase = (double)i/(double)WFCLENGTH*2*M_PI + phase;
                        currentphase *= wavecount;
                        while(currentphase > 2*M_PI)
                            currentphase -= 2*M_PI;
                        tempwaveform[i] += WFRECT(currentphase)*gain;
                    }
                    [s appendString:@"RECT"];
                    break;
                case WFTriangle:
                    for(i=0;i<WFCLENGTH;++i) {
                        double currentphase = (double)i/(double)WFCLENGTH*2*M_PI + phase;
                        currentphase *= wavecount;
                        //currentphase += M_PI_2;
                        while(currentphase > 2*M_PI)
                            currentphase -= 2*M_PI;
                        //currentphase += M_PI_2;
                        tempwaveform[i] += WFTRIANGLE(currentphase)*gain;
                    }
                    [s appendString:@"TRI"];
                    break;
                case WFSawtooth:
                    for(i=0;i<WFCLENGTH;++i) {
                        double currentphase = (double)i/(double)WFCLENGTH*2*M_PI + phase;
                        currentphase *= wavecount;
                        while(currentphase > 2*M_PI)
                            currentphase -= 2*M_PI;
                        tempwaveform[i] += WFSAWTOOTH(currentphase)*gain;
                    }
                    [s appendString:@"SAW"];
                    break;
                default:
                    break;
            }
            [s appendFormat:@" C:%i G:%2.2f P:%2.2f", wavecount, gain, phase];
        }
    }

    for(i=0;i<WFCLENGTH;++i)
        waveform[i] = tempwaveform[i];

    [waveformView constructGraph:self];
    [configField setStringValue:s];
}

- (double)factorForTime:(double)time {
    return [factor doubleValue];
}

- (double)positionForTime:(double)time {
    return [position doubleValue]*(-1.);
}

- (double)amplitudeForTime:(double)t {
    return 1.;
}

- (Float32)amplitudeForPhase:(Float32)phase {
    return waveform[(int)floor(phase/2./M_PI*WFCLENGTH)];
}

- (BOOL)saveWaveformToAIFFAtPath:(NSString *)filepath {
    NSString *folderPath = [[filepath stringByStandardizingPath] stringByDeletingLastPathComponent];
    NSString *fileName = [[filepath stringByStandardizingPath] lastPathComponent];
    
    FSRef folderRef;
    Boolean isDirectory;
    OSStatus err = FSPathMakeRef((UInt8 *)[folderPath UTF8String],
                                 &folderRef,
                                 &isDirectory);
    if (err != noErr || !isDirectory) {
        return NO;
    }
    
    UInt32 theFormatFlags =  kLinearPCMFormatFlagIsFloat
#if BYTE_ORDER == BIG_ENDIAN
        | kLinearPCMFormatFlagIsBigEndian
#endif
        | kLinearPCMFormatFlagIsPacked;
    AudioStreamBasicDescription streamFormat, destFormat;
    streamFormat.mSampleRate = 44100.;		//	the sample rate of the audio stream
    streamFormat.mFormatID = kAudioFormatLinearPCM;			//	the specific encoding type of audio stream
    streamFormat.mFormatFlags = theFormatFlags;		//	flags specific to each format
    streamFormat.mBytesPerPacket = 4;
    streamFormat.mFramesPerPacket = 1;
    streamFormat.mBytesPerFrame = 4;
    streamFormat.mChannelsPerFrame = 1;
    streamFormat.mBitsPerChannel = 32;
    
    destFormat = streamFormat;
    destFormat.mFormatFlags = kLinearPCMFormatFlagIsFloat | kLinearPCMFormatFlagIsBigEndian | kLinearPCMFormatFlagIsPacked;
    
    AudioConverterRef audioConverter;
    err = AudioConverterNew(&streamFormat,
                            &destFormat,
                            &audioConverter);
    if (err != noErr) {
        return NO;
    }
    Float32 outWaveform[WFCLENGTH];
    UInt32 wantedBytes = WFCLENGTH*sizeof(Float32);
    UInt32 writtenBytes = wantedBytes;
    err = AudioConverterConvertBuffer(audioConverter,
                                      WFCLENGTH*sizeof(Float32),
                                      waveform,
                                      &writtenBytes,
                                      outWaveform);
    if (err != noErr || wantedBytes != writtenBytes) {
        return NO;
    }
    AudioConverterDispose(audioConverter);
    
    FSRef fileRef;
    AudioFileID audioFileID;
    err = AudioFileCreate(&folderRef,
                          (CFStringRef)fileName,
                          kAudioFileAIFCType,
                          &destFormat,
                          0,
                          &fileRef,
                          &audioFileID);
    if (err != noErr) {
        return NO;
    }
    
    wantedBytes = WFCLENGTH*sizeof(Float32);
    writtenBytes = wantedBytes;
    err = AudioFileWriteBytes(audioFileID,
                              NO,
                              0,
                              &writtenBytes,
                              outWaveform);
    if (err != noErr || writtenBytes != wantedBytes) {
        return NO;
    }
    
    err = AudioFileClose(audioFileID);
    if (err != noErr) {
        return NO;
    }
    
    return YES;
}

- (IBAction)save:(id)sender {
    NSSavePanel *savePanel = [NSSavePanel savePanel];
    NSArray *fileTypes = [NSArray arrayWithObjects:@"aiff", nil]; //[ @"aiff", nil];
    [savePanel setAllowedFileTypes:fileTypes];
    
    int status = [savePanel runModal];
    
    if (status != NSFileHandlingPanelOKButton) {
        return;
    }
    
    if (![self saveWaveformToAIFFAtPath:[savePanel filename]]) {
        NSBeep();
    }
}

@end
